@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2013 Junjiro R. Okajima
+ * Copyright (C) 2005-2015 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -61,8 +61,11 @@ void au_cpup_attr_nlink(struct inode *inode, int force)
 
 	/*
 	 * 0 can happen in revalidating.
-	 * h_inode->i_mutex is not held, but it is harmless since once i_nlink
-	 * reaches 0, it will never become positive.
+	 * h_inode->i_mutex may not be held here, but it is harmless since once
+	 * i_nlink reaches 0, it will never become positive except O_TMPFILE
+	 * case.
+	 * todo: O_TMPFILE+linkat(AT_SYMLINK_FOLLOW) bypassing aufs may cause
+	 *	 the incorrect link count.
 	 */
 	set_nlink(inode, h_inode->i_nlink);
 
@@ -143,7 +146,8 @@ void au_dtime_revert(struct au_dtime *dt)
 	attr.ia_valid = ATTR_FORCE | ATTR_MTIME | ATTR_MTIME_SET
 		| ATTR_ATIME | ATTR_ATIME_SET;
 
-	err = vfsub_notify_change(&dt->dt_h_path, &attr);
+	/* no delegation since this is a directory */
+	err = vfsub_notify_change(&dt->dt_h_path, &attr, /*delegated*/NULL);
 	if (unlikely(err))
 		pr_warn("restoring timestamps failed(%d). ignored\n", err);
 }
@@ -161,15 +165,19 @@ static noinline_for_stack
 int cpup_iattr(struct dentry *dst, aufs_bindex_t bindex, struct dentry *h_src,
 	       struct au_cpup_reg_attr *h_src_attr)
 {
-	int err, sbits;
+	int err, sbits, icex;
+	unsigned int mnt_flags;
+	unsigned char verbose;
 	struct iattr ia;
 	struct path h_path;
 	struct inode *h_isrc, *h_idst;
 	struct kstat *h_st;
+	struct au_branch *br;
 
 	h_path.dentry = au_h_dptr(dst, bindex);
 	h_idst = h_path.dentry->d_inode;
-	h_path.mnt = au_sbr_mnt(dst->d_sb, bindex);
+	br = au_sbr(dst->d_sb, bindex);
+	h_path.mnt = au_br_mnt(br);
 	h_isrc = h_src->d_inode;
 	ia.ia_valid = ATTR_FORCE | ATTR_UID | ATTR_GID
 		| ATTR_ATIME | ATTR_MTIME
@@ -200,13 +208,21 @@ int cpup_iattr(struct dentry *dst, aufs_bindex_t bindex, struct dentry *h_src,
 		sbits = !!(h_isrc->i_mode & (S_ISUID | S_ISGID));
 		au_cpup_attr_flags(h_idst, h_isrc->i_flags);
 	}
-	err = vfsub_notify_change(&h_path, &ia);
+	/* no delegation since it is just created */
+	err = vfsub_notify_change(&h_path, &ia, /*delegated*/NULL);
 
 	/* is this nfs only? */
 	if (!err && sbits && au_test_nfs(h_path.dentry->d_sb)) {
 		ia.ia_valid = ATTR_FORCE | ATTR_MODE;
 		ia.ia_mode = h_isrc->i_mode;
-		err = vfsub_notify_change(&h_path, &ia);
+		err = vfsub_notify_change(&h_path, &ia, /*delegated*/NULL);
+	}
+
+	icex = br->br_perm & AuBrAttr_ICEX;
+	if (!err) {
+		mnt_flags = au_mntflags(dst->d_sb);
+		verbose = !!au_opt_test(mnt_flags, VERBOSE);
+		err = au_cpup_xattr(h_path.dentry, h_src, icex, verbose);
 	}
 
 	return err;
@@ -303,7 +319,9 @@ static int au_do_copy_file(struct file *dst, struct file *src, loff_t len,
 			ia->ia_file = dst;
 			h_mtx = &file_inode(dst)->i_mutex;
 			mutex_lock_nested(h_mtx, AuLsc_I_CHILD2);
-			err = vfsub_notify_change(&dst->f_path, ia);
+			/* no delegation since it is just created */
+			err = vfsub_notify_change(&dst->f_path, ia,
+						  /*delegated*/NULL);
 			mutex_unlock(h_mtx);
 		}
 	}
@@ -360,20 +378,18 @@ static int au_cp_regular(struct au_cp_generic *cpg)
 		struct dentry *dentry;
 		int force_wr;
 		struct file *file;
-		void *label, *label_file;
+		void *label;
 	} *f, file[] = {
 		{
 			.bindex = cpg->bsrc,
 			.flags = O_RDONLY | O_NOATIME | O_LARGEFILE,
-			.label = &&out,
-			.label_file = &&out_src
+			.label = &&out
 		},
 		{
 			.bindex = cpg->bdst,
 			.flags = O_WRONLY | O_NOATIME | O_LARGEFILE,
 			.force_wr = !!au_ftest_cpup(cpg->flags, RWDST),
-			.label = &&out_src,
-			.label_file = &&out_dst
+			.label = &&out_src
 		}
 	};
 	struct super_block *sb;
@@ -388,18 +404,15 @@ static int au_cp_regular(struct au_cp_generic *cpg)
 		err = PTR_ERR(f->file);
 		if (IS_ERR(f->file))
 			goto *f->label;
-		err = -EINVAL;
-		if (unlikely(!f->file->f_op))
-			goto *f->label_file;
 	}
 
 	/* try stopping to update while we copyup */
 	IMustLock(file[SRC].dentry->d_inode);
 	err = au_copy_file(file[DST].file, file[SRC].file, cpg->len);
 
-out_dst:
 	fput(file[DST].file);
 	au_sbr_put(sb, file[DST].bindex);
+
 out_src:
 	fput(file[SRC].file);
 	au_sbr_put(sb, file[SRC].bindex);
@@ -413,7 +426,7 @@ static int au_do_cpup_regular(struct au_cp_generic *cpg,
 	int err, rerr;
 	loff_t l;
 	struct path h_path;
-	struct inode *h_src_inode;
+	struct inode *h_src_inode, *h_dst_inode;
 
 	err = 0;
 	h_src_inode = au_h_iptr(cpg->dentry->d_inode, cpg->bsrc);
@@ -440,6 +453,13 @@ static int au_do_cpup_regular(struct au_cp_generic *cpg,
 		if (!err && rerr)
 			err = rerr;
 	}
+	if (!err && (h_src_inode->i_state & I_LINKABLE)) {
+		h_path.dentry = au_h_dptr(cpg->dentry, cpg->bdst);
+		h_dst_inode = h_path.dentry->d_inode;
+		spin_lock(&h_dst_inode->i_lock);
+		h_dst_inode->i_state |= I_LINKABLE;
+		spin_unlock(&h_dst_inode->i_lock);
+	}
 
 out:
 	return err;
@@ -488,7 +508,7 @@ int cpup_entry(struct au_cp_generic *cpg, struct dentry *dst_parent,
 	int err;
 	umode_t mode;
 	unsigned int mnt_flags;
-	unsigned char isdir;
+	unsigned char isdir, isreg, force;
 	const unsigned char do_dt = !!au_ftest_cpup(cpg->flags, DTIME);
 	struct au_dtime dt;
 	struct path h_path;
@@ -519,10 +539,12 @@ int cpup_entry(struct au_cp_generic *cpg, struct dentry *dst_parent,
 	}
 	h_path.dentry = h_dst;
 
+	isreg = 0;
 	isdir = 0;
 	mode = h_inode->i_mode;
 	switch (mode & S_IFMT) {
 	case S_IFREG:
+		isreg = 1;
 		err = vfsub_create(h_dir, &h_path, mode | S_IWUSR,
 				   /*want_excl*/true);
 		if (!err)
@@ -562,7 +584,8 @@ int cpup_entry(struct au_cp_generic *cpg, struct dentry *dst_parent,
 	if (!au_opt_test(mnt_flags, UDBA_NONE)
 	    && !isdir
 	    && au_opt_test(mnt_flags, XINO)
-	    && h_inode->i_nlink == 1
+	    && (h_inode->i_nlink == 1
+		|| (h_inode->i_state & I_LINKABLE))
 	    /* todo: unnecessary? */
 	    /* && cpg->dentry->d_inode->i_nlink == 1 */
 	    && cpg->bdst < cpg->bsrc
@@ -570,6 +593,16 @@ int cpup_entry(struct au_cp_generic *cpg, struct dentry *dst_parent,
 		au_xino_write(sb, cpg->bsrc, h_inode->i_ino, /*ino*/0);
 		/* ignore this error */
 
+	if (!err) {
+		force = 0;
+		if (isreg) {
+			force = !!cpg->len;
+			if (cpg->len == -1)
+				force = !!i_size_read(h_inode);
+		}
+		au_fhsm_wrote(sb, cpg->bdst, force);
+	}
+
 	if (do_dt)
 		au_dtime_revert(&dt);
 	return err;
@@ -607,8 +640,9 @@ static int au_do_ren_after_cpup(struct au_cp_generic *cpg, struct path *h_path)
 	h_parent = h_dentry->d_parent; /* dir inode is locked */
 	h_dir = h_parent->d_inode;
 	IMustLock(h_dir);
-	AuDbg("%.*s %.*s\n", AuDLNPair(h_dentry), AuDLNPair(h_path->dentry));
-	err = vfsub_rename(h_dir, h_dentry, h_dir, h_path);
+	AuDbg("%pd %pd\n", h_dentry, h_path->dentry);
+	/* no delegation since it is just created */
+	err = vfsub_rename(h_dir, h_dentry, h_dir, h_path, /*delegated*/NULL);
 	dput(h_path->dentry);
 
 out:
@@ -628,7 +662,7 @@ static int au_cpup_single(struct au_cp_generic *cpg, struct dentry *dst_parent)
 	aufs_bindex_t old_ibstart;
 	unsigned char isdir, plink;
 	struct dentry *h_src, *h_dst, *h_parent;
-	struct inode *dst_inode, *h_dir, *inode;
+	struct inode *dst_inode, *h_dir, *inode, *delegated;
 	struct super_block *sb;
 	struct au_branch *br;
 	/* to reuduce stack size */
@@ -693,11 +727,17 @@ static int au_cpup_single(struct au_cp_generic *cpg, struct dentry *dst_parent)
 			}
 
 			a->h_path.dentry = h_dst;
-			err = vfsub_link(h_src, h_dir, &a->h_path);
+			delegated = NULL;
+			err = vfsub_link(h_src, h_dir, &a->h_path, &delegated);
 			if (!err && au_ftest_cpup(cpg->flags, RENAME))
 				err = au_do_ren_after_cpup(cpg, &a->h_path);
 			if (do_dt)
 				au_dtime_revert(&a->dt);
+			if (unlikely(err == -EWOULDBLOCK)) {
+				pr_warn("cannot retry for NFSv4 delegation"
+					" for an internal link\n");
+				iput(delegated);
+			}
 			dput(h_src);
 			goto out_parent;
 		} else
@@ -747,7 +787,8 @@ static int au_cpup_single(struct au_cp_generic *cpg, struct dentry *dst_parent)
 		goto out_rev;
 
 	if (!isdir
-	    && h_src->d_inode->i_nlink > 1
+	    && (h_src->d_inode->i_nlink > 1
+		|| h_src->d_inode->i_state & I_LINKABLE)
 	    && plink)
 		au_plink_append(inode, cpg->bdst, h_dst);
 
@@ -765,9 +806,11 @@ out_rev:
 	a->h_path.dentry = h_dst;
 	rerr = 0;
 	if (h_dst->d_inode) {
-		if (!isdir)
-			rerr = vfsub_unlink(h_dir, &a->h_path, /*force*/0);
-		else
+		if (!isdir) {
+			/* no delegation since it is just created */
+			rerr = vfsub_unlink(h_dir, &a->h_path,
+					    /*delegated*/NULL, /*force*/0);
+		} else
 			rerr = vfsub_rmdir(h_dir, &a->h_path);
 	}
 	au_dtime_revert(&a->dt);
@@ -782,7 +825,7 @@ out:
 	return err;
 }
 
-#if 0 /* unused */
+#if 0 /* reserved */
 struct au_cpup_single_args {
 	int *errp;
 	struct au_cp_generic *cpg;
@@ -838,7 +881,7 @@ static int au_cpup_sio_test(struct au_pin *pin, umode_t mode)
 	return do_sio;
 }
 
-#if 0 /* unused */
+#if 0 /* reserved */
 int au_sio_cpup_single(struct au_cp_generic *cpg, struct dentry *dst_parent)
 {
 	int err, wkq_err;
@@ -1049,13 +1092,15 @@ static int au_cpup_wh(struct au_cp_generic *cpg, struct file *file)
 
 	dget(wh_dentry);
 	h_path.dentry = wh_dentry;
-	if (!S_ISDIR(wh_dentry->d_inode->i_mode))
-		err = vfsub_unlink(h_parent->d_inode, &h_path, /*force*/0);
-	else
+	if (!d_is_directory(wh_dentry)) {
+		/* no delegation since it is just created */
+		err = vfsub_unlink(h_parent->d_inode, &h_path,
+				   /*delegated*/NULL, /*force*/0);
+	} else
 		err = vfsub_rmdir(h_parent->d_inode, &h_path);
 	if (unlikely(err)) {
-		AuIOErr("failed remove copied-up tmp file %.*s(%d)\n",
-			AuDLNPair(wh_dentry), err);
+		AuIOErr("failed remove copied-up tmp file %pd(%d)\n",
+			wh_dentry, err);
 		err = -EIO;
 	}
 	au_dtime_revert(&dt);
@@ -1087,7 +1132,7 @@ int au_sio_cpup_wh(struct au_cp_generic *cpg, struct file *file)
 {
 	int err, wkq_err;
 	aufs_bindex_t bdst;
-	struct dentry *dentry, *parent, *h_orph, *h_parent, *h_dentry;
+	struct dentry *dentry, *parent, *h_orph, *h_parent;
 	struct inode *dir, *h_dir, *h_tmpdir;
 	struct au_wbr *wbr;
 	struct au_pin wh_pin, *pin_orig;
@@ -1110,10 +1155,6 @@ int au_sio_cpup_wh(struct au_cp_generic *cpg, struct file *file)
 		h_tmpdir = h_orph->d_inode;
 		au_set_h_iptr(dir, bdst, au_igrab(h_tmpdir), /*flags*/0);
 
-		if (file)
-			h_dentry = au_hf_top(file)->f_dentry;
-		else
-			h_dentry = au_h_dptr(dentry, au_dbstart(dentry));
 		mutex_lock_nested(&h_tmpdir->i_mutex, AuLsc_I_PARENT3);
 		/* todo: au_h_open_pre()? */
 
@@ -1223,7 +1264,7 @@ out:
 
 static int au_cpup_dir(struct dentry *dentry, aufs_bindex_t bdst,
 		       struct au_pin *pin,
-		       struct dentry *h_parent __maybe_unused ,
+		       struct dentry *h_parent __maybe_unused,
 		       void *arg __maybe_unused)
 {
 	struct au_cp_generic cpg = {
