@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2014 Junjiro R. Okajima
+ * Copyright (C) 2005-2015 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -388,7 +388,7 @@ static struct dentry *lookup_wlock_by_name(char *name, unsigned int nlen,
 			au_digen_dec(d);
 		else
 			goto cont_unlock;
-		if (d_count(d)) {
+		if (au_dcount(d) > 0) {
 			dentry = dget_dlock(d);
 			spin_unlock(&d->d_lock);
 			break;
@@ -586,7 +586,7 @@ int au_hnotify(struct inode *h_dir, struct au_hnotify *hnotify, u32 mask,
 		au_fset_hnjob(flags[AuHn_CHILD], MNTPNT);
 		/*FALLTHROUGH*/
 	case FS_CREATE:
-		AuDebugOn(!h_child_name || !h_child_inode);
+		AuDebugOn(!h_child_name);
 		break;
 
 	case FS_DELETE:
