@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2013 Junjiro R. Okajima
+ * Copyright (C) 2005-2015 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -32,7 +32,6 @@ enum { AuPARENT, AuCHILD, AuParentChild };
 #define AuRen_MNT_WRITE	(1 << 4)
 #define AuRen_DT_DSTDIR	(1 << 5)
 #define AuRen_DIROPQ	(1 << 6)
-#define AuRen_CPUP	(1 << 7)
 #define au_ftest_ren(flags, name)	((flags) & AuRen_##name)
 #define au_fset_ren(flags, name) \
 	do { (flags) |= AuRen_##name; } while (0)
@@ -111,52 +110,48 @@ static void au_ren_rev_diropq(int err, struct au_ren_args *a)
 	au_hn_imtx_unlock(a->src_hinode);
 	au_set_dbdiropq(a->src_dentry, a->src_bdiropq);
 	if (rerr)
-		RevertFailure("remove diropq %.*s", AuDLNPair(a->src_dentry));
+		RevertFailure("remove diropq %pd", a->src_dentry);
 }
 
 static void au_ren_rev_rename(int err, struct au_ren_args *a)
 {
 	int rerr;
+	struct inode *delegated;
 
 	a->h_path.dentry = vfsub_lkup_one(&a->src_dentry->d_name,
 					  a->src_h_parent);
 	rerr = PTR_ERR(a->h_path.dentry);
 	if (IS_ERR(a->h_path.dentry)) {
-		RevertFailure("lkup one %.*s", AuDLNPair(a->src_dentry));
+		RevertFailure("lkup one %pd", a->src_dentry);
 		return;
 	}
 
+	delegated = NULL;
 	rerr = vfsub_rename(a->dst_h_dir,
 			    au_h_dptr(a->src_dentry, a->btgt),
-			    a->src_h_dir, &a->h_path);
+			    a->src_h_dir, &a->h_path, &delegated);
+	if (unlikely(rerr == -EWOULDBLOCK)) {
+		pr_warn("cannot retry for NFSv4 delegation"
+			" for an internal rename\n");
+		iput(delegated);
+	}
 	d_drop(a->h_path.dentry);
 	dput(a->h_path.dentry);
 	/* au_set_h_dptr(a->src_dentry, a->btgt, NULL); */
 	if (rerr)
-		RevertFailure("rename %.*s", AuDLNPair(a->src_dentry));
-}
-
-static void au_ren_rev_cpup(int err, struct au_ren_args *a)
-{
-	int rerr;
-
-	a->h_path.dentry = a->dst_h_dentry;
-	rerr = vfsub_unlink(a->dst_h_dir, &a->h_path, /*force*/0);
-	au_set_h_dptr(a->src_dentry, a->btgt, NULL);
-	au_set_dbstart(a->src_dentry, a->src_bstart);
-	if (rerr)
-		RevertFailure("unlink %.*s", AuDLNPair(a->dst_h_dentry));
+		RevertFailure("rename %pd", a->src_dentry);
 }
 
 static void au_ren_rev_whtmp(int err, struct au_ren_args *a)
 {
 	int rerr;
+	struct inode *delegated;
 
 	a->h_path.dentry = vfsub_lkup_one(&a->dst_dentry->d_name,
 					  a->dst_h_parent);
 	rerr = PTR_ERR(a->h_path.dentry);
 	if (IS_ERR(a->h_path.dentry)) {
-		RevertFailure("lkup one %.*s", AuDLNPair(a->dst_dentry));
+		RevertFailure("lkup one %pd", a->dst_dentry);
 		return;
 	}
 	if (a->h_path.dentry->d_inode) {
@@ -165,13 +160,20 @@ static void au_ren_rev_whtmp(int err, struct au_ren_args *a)
 		return;
 	}
 
-	rerr = vfsub_rename(a->dst_h_dir, a->h_dst, a->dst_h_dir, &a->h_path);
+	delegated = NULL;
+	rerr = vfsub_rename(a->dst_h_dir, a->h_dst, a->dst_h_dir, &a->h_path,
+			    &delegated);
+	if (unlikely(rerr == -EWOULDBLOCK)) {
+		pr_warn("cannot retry for NFSv4 delegation"
+			" for an internal rename\n");
+		iput(delegated);
+	}
 	d_drop(a->h_path.dentry);
 	dput(a->h_path.dentry);
 	if (!rerr)
 		au_set_h_dptr(a->dst_dentry, a->btgt, dget(a->h_dst));
 	else
-		RevertFailure("rename %.*s", AuDLNPair(a->h_dst));
+		RevertFailure("rename %pd", a->h_dst);
 }
 
 static void au_ren_rev_whsrc(int err, struct au_ren_args *a)
@@ -182,7 +184,7 @@ static void au_ren_rev_whsrc(int err, struct au_ren_args *a)
 	rerr = au_wh_unlink_dentry(a->src_h_dir, &a->h_path, a->src_dentry);
 	au_set_dbwh(a->src_dentry, a->src_bwh);
 	if (rerr)
-		RevertFailure("unlink %.*s", AuDLNPair(a->src_wh_dentry));
+		RevertFailure("unlink %pd", a->src_wh_dentry);
 }
 #undef RevertFailure
 
@@ -197,6 +199,7 @@ static int au_ren_or_cpup(struct au_ren_args *a)
 {
 	int err;
 	struct dentry *d;
+	struct inode *delegated;
 
 	d = a->src_dentry;
 	if (au_dbstart(d) == a->btgt) {
@@ -205,8 +208,14 @@ static int au_ren_or_cpup(struct au_ren_args *a)
 		    && au_dbdiropq(d) == a->btgt)
 			au_fclr_ren(a->flags, DIROPQ);
 		AuDebugOn(au_dbstart(d) != a->btgt);
+		delegated = NULL;
 		err = vfsub_rename(a->src_h_dir, au_h_dptr(d, a->btgt),
-				   a->dst_h_dir, &a->h_path);
+				   a->dst_h_dir, &a->h_path, &delegated);
+		if (unlikely(err == -EWOULDBLOCK)) {
+			pr_warn("cannot retry for NFSv4 delegation"
+				" for an internal rename\n");
+			iput(delegated);
+		}
 	} else
 		BUG();
 
@@ -230,8 +239,8 @@ static int au_ren_del_whtmp(struct au_ren_args *a)
 	    || au_test_fs_remote(a->h_dst->d_sb)) {
 		err = au_whtmp_rmdir(dir, a->btgt, a->h_dst, &a->whlist);
 		if (unlikely(err))
-			pr_warn("failed removing whtmp dir %.*s (%d), "
-				"ignored.\n", AuDLNPair(a->h_dst), err);
+			pr_warn("failed removing whtmp dir %pd (%d), "
+				"ignored.\n", a->h_dst, err);
 	} else {
 		au_nhash_wh_free(&a->thargs->whlist);
 		a->thargs->whlist = a->whlist;
@@ -258,7 +267,8 @@ static int au_ren_diropq(struct au_ren_args *a)
 	au_hn_imtx_unlock(a->src_hinode);
 	if (IS_ERR(diropq))
 		err = PTR_ERR(diropq);
-	dput(diropq);
+	else
+		dput(diropq);
 
 	return err;
 }
@@ -358,6 +368,7 @@ static int do_rename(struct au_ren_args *a)
 	if (a->thargs)
 		au_ren_del_whtmp(a); /* ignore this error */
 
+	au_fhsm_wrote(a->src_dentry->d_sb, a->btgt, /*force*/0);
 	err = 0;
 	goto out_success;
 
@@ -365,10 +376,7 @@ out_diropq:
 	if (au_ftest_ren(a->flags, DIROPQ))
 		au_ren_rev_diropq(err, a);
 out_rename:
-	if (!au_ftest_ren(a->flags, CPUP))
-		au_ren_rev_rename(err, a);
-	else
-		au_ren_rev_cpup(err, a);
+	au_ren_rev_rename(err, a);
 	dput(a->h_dst);
 out_whtmp:
 	if (a->thargs)
@@ -627,8 +635,7 @@ static void au_ren_refresh_dir(struct au_ren_args *a)
 		au_cpup_attr_nlink(dir, /*force*/1);
 	}
 
-	if (au_ibstart(dir) == a->btgt)
-		au_cpup_attr_timesizes(dir);
+	au_dir_ts(dir, a->btgt);
 
 	if (au_ftest_ren(a->flags, ISSAMEDIR))
 		return;
@@ -637,8 +644,7 @@ static void au_ren_refresh_dir(struct au_ren_args *a)
 	dir->i_version++;
 	if (au_ftest_ren(a->flags, ISDIR))
 		au_cpup_attr_nlink(dir, /*force*/1);
-	if (au_ibstart(dir) == a->btgt)
-		au_cpup_attr_timesizes(dir);
+	au_dir_ts(dir, a->btgt);
 }
 
 static void au_ren_refresh(struct au_ren_args *a)
@@ -804,7 +810,7 @@ int aufs_rename(struct inode *_src_dir, struct dentry *_src_dentry,
 	/* reduce stack space */
 	struct au_ren_args *a;
 
-	AuDbg("%.*s, %.*s\n", AuDLNPair(_src_dentry), AuDLNPair(_dst_dentry));
+	AuDbg("%pd, %pd\n", _src_dentry, _dst_dentry);
 	IMustLock(_src_dir);
 	IMustLock(_dst_dir);
 
@@ -829,9 +835,10 @@ int aufs_rename(struct inode *_src_dir, struct dentry *_src_dentry,
 
 	err = -ENOTDIR;
 	flags = AuLock_FLUSH | AuLock_NOPLM | AuLock_GEN;
-	if (S_ISDIR(a->src_inode->i_mode)) {
+	if (d_is_directory(a->src_dentry)) {
 		au_fset_ren(a->flags, ISDIR);
-		if (unlikely(a->dst_inode && !S_ISDIR(a->dst_inode->i_mode)))
+		if (unlikely(d_is_positive(a->dst_dentry)
+			     && !d_is_directory(a->dst_dentry)))
 			goto out_free;
 		err = aufs_read_and_write_lock2(a->dst_dentry, a->src_dentry,
 						AuLock_DIR | flags);
