@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2013 Junjiro R. Okajima
+ * Copyright (C) 2005-2015 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,26 +27,23 @@ unsigned int aufs_poll(struct file *file, poll_table *wait)
 	unsigned int mask;
 	int err;
 	struct file *h_file;
-	struct dentry *dentry;
 	struct super_block *sb;
 
 	/* We should pretend an error happened. */
 	mask = POLLERR /* | POLLIN | POLLOUT */;
-	dentry = file->f_dentry;
-	sb = dentry->d_sb;
+	sb = file->f_dentry->d_sb;
 	si_read_lock(sb, AuLock_FLUSH | AuLock_NOPLMW);
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/0);
-	if (unlikely(err))
+
+	h_file = au_read_pre(file, /*keep_fi*/0);
+	err = PTR_ERR(h_file);
+	if (IS_ERR(h_file))
 		goto out;
 
 	/* it is not an error if h_file has no operation */
 	mask = DEFAULT_POLLMASK;
-	h_file = au_hf_top(file);
-	if (h_file->f_op && h_file->f_op->poll)
+	if (h_file->f_op->poll)
 		mask = h_file->f_op->poll(h_file, wait);
-
-	di_read_unlock(dentry, AuLock_IR);
-	fi_read_unlock(file);
+	fput(h_file); /* instead of au_read_post() */
 
 out:
 	si_read_unlock(sb);
